@@ -7,7 +7,7 @@
  * slab size is always 1MB, since that's the maximum item size allowed by the
  * memcached protocol.
  *
- * $Id: slabs.c 591 2007-07-09 14:28:54Z plindner $
+ * $Id$
  */
 #include "memcached.h"
 #include <sys/stat.h>
@@ -25,7 +25,7 @@
 #define POWER_SMALLEST 1
 #define POWER_LARGEST  200
 #define POWER_BLOCK 1048576
-#define CHUNK_ALIGN_BYTES (sizeof(void *))
+#define CHUNK_ALIGN_BYTES 8
 #define DONT_PREALLOC_SLABS
 
 /* powers-of-N allocation structures */
@@ -54,10 +54,15 @@ static size_t mem_limit = 0;
 static size_t mem_malloced = 0;
 static int power_largest;
 
+static void *mem_base = NULL;
+static void *mem_current = NULL;
+static size_t mem_avail = 0;
+
 /*
  * Forward Declarations
  */
 static int do_slabs_newslab(const unsigned int id);
+static void *memory_allocate(size_t size);
 
 #ifndef DONT_PREALLOC_SLABS
 /* Preallocate as many slab pages as possible (called from slabs_init)
@@ -92,7 +97,7 @@ unsigned int slabs_clsid(const size_t size) {
  * Determines the chunk sizes and initializes the slab class descriptors
  * accordingly.
  */
-void slabs_init(const size_t limit, const double factor) {
+void slabs_init(const size_t limit, const double factor, const bool prealloc) {
     int i = POWER_SMALLEST - 1;
     unsigned int size = sizeof(item) + settings.chunk_size;
 
@@ -101,6 +106,19 @@ void slabs_init(const size_t limit, const double factor) {
         size = 128;
 
     mem_limit = limit;
+
+    if (prealloc) {
+        /* Allocate everything in a big chunk with malloc */
+        mem_base = malloc(mem_limit);
+        if (mem_base != NULL) {
+            mem_current = mem_base;
+            mem_avail = mem_limit;
+        } else {
+            fprintf(stderr, "Warning: Failed to allocate requested memory in"
+                    " one large chunk.\nWill allocate in smaller chunks\n");
+        }
+    }
+
     memset(slabclass, 0, sizeof(slabclass));
 
     while (++i < POWER_LARGEST && size <= POWER_BLOCK / 2) {
@@ -182,13 +200,13 @@ static int do_slabs_newslab(const unsigned int id) {
 #endif
     char *ptr;
 
-    if (mem_limit && mem_malloced + len > mem_limit && p->slabs > 0)
+    if ((mem_limit && mem_malloced + len > mem_limit && p->slabs > 0) ||
+        (grow_slab_list(id) == 0) ||
+        ((ptr = memory_allocate((size_t)len)) == 0)) {
+
+        MEMCACHED_SLABS_SLABCLASS_ALLOCATE_FAILED(id);
         return 0;
-
-    if (grow_slab_list(id) == 0) return 0;
-
-    ptr = malloc((size_t)len);
-    if (ptr == 0) return 0;
+    }
 
     memset(ptr, 0, (size_t)len);
     p->end_page_ptr = ptr;
@@ -196,52 +214,65 @@ static int do_slabs_newslab(const unsigned int id) {
 
     p->slab_list[p->slabs++] = ptr;
     mem_malloced += len;
+
+    MEMCACHED_SLABS_SLABCLASS_ALLOCATE(id);
     return 1;
 }
 
 /*@null@*/
-void *do_slabs_alloc(const size_t size) {
+void *do_slabs_alloc(const size_t size, unsigned int id) {
     slabclass_t *p;
+    void *ret = NULL;
 
-    unsigned int id = slabs_clsid(size);
-    if (id < POWER_SMALLEST || id > power_largest)
+    if (id < POWER_SMALLEST || id > power_largest) {
+        MEMCACHED_SLABS_ALLOCATE_FAILED(size, 0);
         return NULL;
+    }
 
     p = &slabclass[id];
     assert(p->sl_curr == 0 || ((item *)p->slots[p->sl_curr - 1])->slabs_clsid == 0);
 
 #ifdef USE_SYSTEM_MALLOC
-    if (mem_limit && mem_malloced + size > mem_limit)
+    if (mem_limit && mem_malloced + size > mem_limit) {
+        MEMCACHED_SLABS_ALLOCATE_FAILED(size, id);
         return 0;
+    }
     mem_malloced += size;
-    return malloc(size);
+    ret = malloc(size);
+    MEMCACHED_SLABS_ALLOCATE(size, id, 0, ret);
+    return ret;
 #endif
 
     /* fail unless we have space at the end of a recently allocated page,
        we have something on our freelist, or we could allocate a new page */
-    if (! (p->end_page_ptr != 0 || p->sl_curr != 0 || do_slabs_newslab(id) != 0))
-        return 0;
-
-    /* return off our freelist, if we have one */
-    if (p->sl_curr != 0)
-        return p->slots[--p->sl_curr];
-
-    /* if we recently allocated a whole page, return from that */
-    if (p->end_page_ptr) {
-        void *ptr = p->end_page_ptr;
+    if (! (p->end_page_ptr != 0 || p->sl_curr != 0 ||
+           do_slabs_newslab(id) != 0)) {
+        /* We don't have more memory available */
+        ret = NULL;
+    } else if (p->sl_curr != 0) {
+        /* return off our freelist */
+        ret = p->slots[--p->sl_curr];
+    } else {
+        /* if we recently allocated a whole page, return from that */
+        assert(p->end_page_ptr != NULL);
+        ret = p->end_page_ptr;
         if (--p->end_page_free != 0) {
             p->end_page_ptr += p->size;
         } else {
             p->end_page_ptr = 0;
         }
-        return ptr;
     }
 
-    return NULL;  /* shouldn't ever get here */
+    if (ret) {
+        MEMCACHED_SLABS_ALLOCATE(size, id, p->size, ret);
+    } else {
+        MEMCACHED_SLABS_ALLOCATE_FAILED(size, id);
+    }
+
+    return ret;
 }
 
-void do_slabs_free(void *ptr, const size_t size) {
-    unsigned char id = slabs_clsid(size);
+void do_slabs_free(void *ptr, const size_t size, unsigned int id) {
     slabclass_t *p;
 
     assert(((item *)ptr)->slabs_clsid == 0);
@@ -249,6 +280,7 @@ void do_slabs_free(void *ptr, const size_t size) {
     if (id < POWER_SMALLEST || id > power_largest)
         return;
 
+    MEMCACHED_SLABS_FREE(size, id, ptr);
     p = &slabclass[id];
 
 #ifdef USE_SYSTEM_MALLOC
@@ -374,3 +406,32 @@ int do_slabs_reassign(unsigned char srcid, unsigned char dstid) {
     return 1;
 }
 #endif
+
+static void *memory_allocate(size_t size) {
+    void *ret;
+
+    if (mem_base == NULL) {
+        /* We are not using a preallocated large memory chunk */
+        ret = malloc(size);
+    } else {
+        ret = mem_current;
+
+        if (size > mem_avail) {
+            return NULL;
+        }
+
+        /* mem_current pointer _must_ be aligned!!! */
+        if (size % CHUNK_ALIGN_BYTES) {
+            size += CHUNK_ALIGN_BYTES - (size % CHUNK_ALIGN_BYTES);
+        }
+
+        mem_current += size;
+        if (size < mem_avail) {
+            mem_avail -= size;
+        } else {
+            mem_avail = 0;
+        }
+    }
+
+    return ret;
+}
