@@ -47,6 +47,9 @@ struct conn_queue {
 /* Lock for connection freelist */
 static pthread_mutex_t conn_lock;
 
+/* Lock for alternative item suffix freelist */
+static pthread_mutex_t suffix_lock;
+
 /* Lock for cache operations (item_*, assoc_*) */
 static pthread_mutex_t cache_lock;
 
@@ -235,8 +238,8 @@ conn *mt_conn_from_freelist() {
  *
  * Returns 0 on success, 1 if the structure couldn't be added.
  */
-int mt_conn_add_to_freelist(conn *c) {
-    int result;
+bool mt_conn_add_to_freelist(conn *c) {
+    bool result;
 
     pthread_mutex_lock(&conn_lock);
     result = do_conn_add_to_freelist(c);
@@ -245,6 +248,36 @@ int mt_conn_add_to_freelist(conn *c) {
     return result;
 }
 
+/*
+ * Pulls a suffix buffer from the freelist, if one is available.
+ */
+char *mt_suffix_from_freelist() {
+    char *s;
+
+    pthread_mutex_lock(&suffix_lock);
+    s = do_suffix_from_freelist();
+    pthread_mutex_unlock(&suffix_lock);
+
+    return s;
+}
+
+
+/*
+ * Adds a suffix buffer to the freelist.
+ *
+ * Returns 0 on success, 1 if the buffer couldn't be added.
+ */
+bool mt_suffix_add_to_freelist(char *s) {
+    bool result;
+
+    pthread_mutex_lock(&suffix_lock);
+    result = do_suffix_add_to_freelist(s);
+    pthread_mutex_unlock(&suffix_lock);
+
+    return result;
+}
+
+
 /****************************** LIBEVENT THREADS *****************************/
 
 /*
@@ -310,7 +343,7 @@ static void thread_libevent_process(int fd, short which, void *arg) {
     if (NULL != item) {
         conn *c = conn_new(item->sfd, item->init_state, item->event_flags,
                            item->read_buffer_size, item->is_udp, me->base);
-        if (!c) {
+        if (c == NULL) {
             if (item->is_udp) {
                 fprintf(stderr, "Can't listen for events on UDP socket\n");
                 exit(1);
@@ -348,6 +381,8 @@ void dispatch_conn_new(int sfd, int init_state, int event_flags,
     item->is_udp = is_udp;
 
     cq_push(&threads[thread].new_conn_queue, item);
+
+    MEMCACHED_CONN_DISPATCH(sfd, threads[thread].thread_id);
     if (write(threads[thread].notify_send_fd, "", 1) != 1) {
         perror("Writing to thread notify pipe");
     }
@@ -462,11 +497,12 @@ char *mt_defer_delete(item *item, time_t exptime) {
 /*
  * Does arithmetic on a numeric item value.
  */
-char *mt_add_delta(item *item, int incr, unsigned int delta, char *buf) {
+char *mt_add_delta(conn *c, item *item, int incr, const int64_t delta,
+                   char *buf) {
     char *ret;
 
     pthread_mutex_lock(&cache_lock);
-    ret = do_add_delta(item, incr, delta, buf);
+    ret = do_add_delta(c, item, incr, delta, buf);
     pthread_mutex_unlock(&cache_lock);
     return ret;
 }
@@ -492,6 +528,42 @@ void mt_item_flush_expired() {
     pthread_mutex_unlock(&cache_lock);
 }
 
+/*
+ * Dumps part of the cache
+ */
+char *mt_item_cachedump(unsigned int slabs_clsid, unsigned int limit, unsigned int *bytes) {
+    char *ret;
+
+    pthread_mutex_lock(&cache_lock);
+    ret = do_item_cachedump(slabs_clsid, limit, bytes);
+    pthread_mutex_unlock(&cache_lock);
+    return ret;
+}
+
+/*
+ * Dumps statistics about slab classes
+ */
+char *mt_item_stats(int *bytes) {
+    char *ret;
+
+    pthread_mutex_lock(&cache_lock);
+    ret = do_item_stats(bytes);
+    pthread_mutex_unlock(&cache_lock);
+    return ret;
+}
+
+/*
+ * Dumps a list of objects of each size in 32-byte increments
+ */
+char *mt_item_stats_sizes(int *bytes) {
+    char *ret;
+
+    pthread_mutex_lock(&cache_lock);
+    ret = do_item_stats_sizes(bytes);
+    pthread_mutex_unlock(&cache_lock);
+    return ret;
+}
+
 /****************************** HASHTABLE MODULE *****************************/
 
 void mt_assoc_move_next_bucket() {
@@ -502,18 +574,18 @@ void mt_assoc_move_next_bucket() {
 
 /******************************* SLAB ALLOCATOR ******************************/
 
-void *mt_slabs_alloc(size_t size) {
+void *mt_slabs_alloc(size_t size, unsigned int id) {
     void *ret;
 
     pthread_mutex_lock(&slabs_lock);
-    ret = do_slabs_alloc(size);
+    ret = do_slabs_alloc(size, id);
     pthread_mutex_unlock(&slabs_lock);
     return ret;
 }
 
-void mt_slabs_free(void *ptr, size_t size) {
+void mt_slabs_free(void *ptr, size_t size, unsigned int id) {
     pthread_mutex_lock(&slabs_lock);
-    do_slabs_free(ptr, size);
+    do_slabs_free(ptr, size, id);
     pthread_mutex_unlock(&slabs_lock);
 }
 
@@ -596,7 +668,7 @@ void thread_init(int nthreads, struct event_base *main_base) {
 
     /* Wait for all the threads to set themselves up before returning. */
     pthread_mutex_lock(&init_lock);
-    init_count++; // main thread
+    init_count++; /* main thread */
     while (init_count < nthreads) {
         pthread_cond_wait(&init_cond, &init_lock);
     }
